import { defineNuxtModule, addVitePlugin, installModule } from '@nuxt/kit';
import { walk } from 'estree-walker';
import MagicString from 'magic-string';

const PLUGIN_NAME = "nuxt:vitest:mock-transform";
const HELPER_MOCK_IMPORT = "mockNuxtImport";
const HELPER_MOCK_COMPONENT = "mockComponent";
const HELPERS_NAME = [HELPER_MOCK_IMPORT, HELPER_MOCK_COMPONENT];
const mockTransform = defineNuxtModule({
  meta: {
    name: PLUGIN_NAME
  },
  setup(_, nuxt) {
    let imports = [];
    let components = [];
    nuxt.hook("imports:extend", (_2) => {
      imports = imports.concat(_2);
    });
    nuxt.hook("imports:sources", (_2) => {
      imports = imports.concat(
        // cast presets to imports
        _2.filter((item) => item.from === "#app").flatMap(
          (item) => item.imports.flatMap((name) => {
            return name.toString().startsWith("use") ? {
              name,
              as: name,
              from: item.from
            } : [];
          })
        )
      );
    });
    nuxt.hook("components:extend", (_2) => {
      components = _2;
    });
    addVitePlugin({
      name: PLUGIN_NAME,
      enforce: "post",
      transform: {
        order: "post",
        handler(code, id) {
          if (!HELPERS_NAME.some((n) => code.includes(n)))
            return;
          if (id.includes("/node_modules/"))
            return;
          let ast;
          try {
            ast = this.parse(code, {
              sourceType: "module",
              ecmaVersion: "latest",
              ranges: true
            });
          } catch (e) {
            return;
          }
          let insertionPoint = 0;
          let hasViImport = false;
          const s = new MagicString(code);
          const mocksImport = [];
          const mocksComponent = [];
          walk(ast, {
            enter: (node) => {
              if (node.type === "ImportDeclaration" && node.source.value === "vitest" && !hasViImport) {
                if (node.specifiers.find(
                  (i) => i.type === "ImportSpecifier" && i.imported.name === "vi"
                )) {
                  insertionPoint = node.range[1];
                  hasViImport = true;
                }
                return;
              }
              if (node.type !== "CallExpression")
                return;
              const call = node;
              if (call.callee.type === "Identifier" && call.callee.name === HELPER_MOCK_IMPORT) {
                if (call.arguments.length !== 2) {
                  return this.error(
                    new Error(
                      `${HELPER_MOCK_IMPORT}() should have exactly 2 arguments`
                    ),
                    call.range[0]
                  );
                }
                if (call.arguments[0].type !== "Literal") {
                  return this.error(
                    new Error(
                      `The first argument of ${HELPER_MOCK_IMPORT}() must be a string literal`
                    ),
                    call.arguments[0].range[0]
                  );
                }
                const name = call.arguments[0].value;
                const importItem = imports.find((_2) => name === (_2.as || _2.name));
                if (!importItem) {
                  return this.error(`Cannot find import "${name}" to mock`);
                }
                s.overwrite(call.range[0], call.range[1], "");
                mocksImport.push({
                  name,
                  import: importItem,
                  factory: code.slice(
                    call.arguments[1].range[0],
                    call.arguments[1].range[1]
                  )
                });
              }
              if (call.callee.type === "Identifier" && call.callee.name === HELPER_MOCK_COMPONENT) {
                if (call.arguments.length !== 2) {
                  return this.error(
                    new Error(
                      `${HELPER_MOCK_COMPONENT}() should have exactly 2 arguments`
                    ),
                    call.range[0]
                  );
                }
                if (call.arguments[0].type !== "Literal") {
                  return this.error(
                    new Error(
                      `The first argument of ${HELPER_MOCK_COMPONENT}() must be a string literal`
                    ),
                    call.arguments[0].range[0]
                  );
                }
                const pathOrName = call.arguments[0].value;
                const component = components.find(
                  (_2) => _2.pascalName === pathOrName || _2.kebabName === pathOrName
                );
                const path = component?.filePath || pathOrName;
                s.overwrite(call.range[0], call.range[1], "");
                mocksComponent.push({
                  path,
                  factory: code.slice(
                    call.arguments[1].range[0],
                    call.arguments[1].range[1]
                  )
                });
              }
            }
          });
          if (mocksImport.length === 0 && mocksComponent.length === 0)
            return;
          const mockLines = [];
          if (mocksImport.length) {
            const mockImportMap = /* @__PURE__ */ new Map();
            for (const mock of mocksImport) {
              if (!mockImportMap.has(mock.import.from)) {
                mockImportMap.set(mock.import.from, []);
              }
              mockImportMap.get(mock.import.from).push(mock);
            }
            mockLines.push(
              ...Array.from(mockImportMap.entries()).flatMap(
                ([from, mocks]) => {
                  const lines = [
                    `vi.mock(${JSON.stringify(
                      from
                    )}, async (importOriginal) => {`,
                    `  const mod = { ...await importOriginal() }`
                  ];
                  for (const mock of mocks) {
                    lines.push(
                      `  mod[${JSON.stringify(mock.name)}] = await (${mock.factory})()`
                    );
                  }
                  lines.push(`  return mod`);
                  lines.push(`})`);
                  return lines;
                }
              )
            );
          }
          if (mocksComponent.length) {
            mockLines.push(
              ...mocksComponent.flatMap((mock) => {
                return [
                  `vi.mock(${JSON.stringify(mock.path)}, async () => {`,
                  `  const factory = (${mock.factory});`,
                  `  const result = typeof factory === 'function' ? await factory() : await factory`,
                  `  return 'default' in result ? result : { default: result }`,
                  "})"
                ];
              })
            );
          }
          if (!mockLines.length)
            return;
          if (!hasViImport)
            mockLines.unshift(`import {vi} from "vitest";`);
          s.appendLeft(insertionPoint, "\n" + mockLines.join("\n") + ";\n");
          return {
            code: s.toString(),
            map: s.generateMap()
          };
        }
      }
    });
  }
});

const module = defineNuxtModule({
  meta: {
    name: "vitest-env"
  },
  async setup() {
    await installModule(mockTransform);
  }
});

export { module as default };
